//*******************************************************
//	FM channel Radio
//	FM tuner module : SPK-FM-301 / SPK-TFM-1010
//
//	MCU:	ATmega8
//			hfuse:C1h	EEPROMی
//			lfuse:A4h	CR 8MHz BOD 2.7V
//
//	Compiler:	AVR-GCC		4.3.0
//				AVR-LIBC	1.6.2
//				binutils	2.18
//
//	2009.8.17	KAWAKAMI Yukio
//*******************************************************

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<avr/eeprom.h>
#include	<util/twi.h>
#include	<string.h>

#include	"FK1.h"
#include	"AR1000.h"

#define		CPU_CLOCK	8000000L	// CPUNbN 8MHz
#define		CH_N		4			// `l

#define		SLEEP_MIN	30	// X[v[hœdOFF鎞()

#define		R3_VOLUME	R3_VOLUMN_4		// AR1000 o̓{[

#define		RSSI_MIN_LEVEL	7	// Lȓdg̃x

#define		LOW_BATT	200		// [obe[d 2.00V

#ifdef PRR	// ATmega88/168
#define		ADC_REF	110		// A/Dt@Xd 1.1V
#else		// ATmega8
#define		ADC_REF	256		// A/Dt@Xd 2.56V
#endif

#define		R8_VALUE	3300L	// R R8̒l
#define		R9_VALUE	10000L	// R R9̒l


ushort	R1Default;

volatile uchar	Timer;
volatile uchar	FlashTimer;
volatile uchar	TwiTimer;
volatile uchar	FlashLED;
volatile uchar	FMch;
volatile uchar	SwFlag;
volatile uchar	LEDdata;
volatile uchar	SleepFlag;
volatile uchar	SleepTimerDS;
volatile uchar	SleepTimerS;
volatile uchar	SleepTimerM;
volatile ushort	BattVolt;


//*******************************************************
//	^C}[Qr`rv荞
//	1MHz / 64 / 156 = 100Hz
//*******************************************************
#ifndef TIMER2_COMPA_vect
#define	TIMER2_COMPA_vect	TIMER2_COMP_vect
#endif

ISR(TIMER2_COMPA_vect)
{
	uchar	d;

	// e^C}
	if (Timer)	--Timer;
	if (FlashTimer)	--FlashTimer;
	if (TwiTimer)	--TwiTimer;
	if (SleepFlag){
		if (++SleepTimerDS >= 100){
			SleepTimerDS = 0;
			if (++SleepTimerS >= 60){
				SleepTimerS = 0;
				++SleepTimerM;
			}
		}
	}

	// XCb`͒l SwFlagɃZbg
	SwFlag = 0;
	d = ~PIND;

	if (d & (1<<SW0))		SwFlag |= SW_POWER;
	if (d & (1<<SW1))		SwFlag |= SW_1;
	if (d & (1<<SW2))		SwFlag |= SW_2;
	if (d & (1<<SW3))		SwFlag |= SW_3;
	if (d & (1<<SW4))		SwFlag |= SW_4;

	// LED_ŏ
	if (FlashLED){
		if (FlashTimer == 0){
			LEDdata = 1 << FlashLED;
			FlashTimer = 50;
		} else if (FlashTimer == 25){
			LEDdata = 0;
		}
	}

	if (LEDdata & SW_1)		PORTB |= (1<<LED1); else PORTB &= ~(1<<LED1);
	if (LEDdata & SW_2)		PORTB |= (1<<LED2); else PORTB &= ~(1<<LED2);
	if (LEDdata & SW_3)		PORTB |= (1<<LED3); else PORTB &= ~(1<<LED3);
	if (LEDdata & SW_4)		PORTB |= (1<<LED4); else PORTB &= ~(1<<LED4);

	// A/DϊJn vXP[ 8MHz/64 = 125kHz
	ADCSRA = (1<<ADEN)|(1<<ADSC)|(1<<ADIE)|(1<<ADPS1)|(1<<ADPS2);
}


//*******************************************************
//	A/Dϊ̊荞݃[`
//*******************************************************
ISR(ADC_vect) {

	BattVolt = (uint)ADCL;
	BattVolt |= (((uint)ADCH & 3) << 8);
	ADCSRA &= ~(1<<ADEN);	// A/D~
}


//*******************************************************
//	C[`
//*******************************************************
int main (void) {
	uint	freq;
	uchar	off_count;

	cli();	// 荞݋֎~

	// |[gݒ
	PORTB = (1<<PB7)|(1<<PB6)|(1<<SCK)|(1<<MISO);
	DDRB = (1<<LED4)|(1<<LED3)|(1<<LED2)|(1<<LED1);

	PORTD = (1<<SW4)|(1<<SW3)|(1<<SW2)|(1<<SW1)|(1<<SW0)|(1<<TXD)|(1<<RXD);
	DDRD = (1<<POWER);

	PORTC = (1<<PC3)|(1<<PC2)|(1<<PC1);
	DDRC = 0;

#ifdef PRR	// ATmega88/168
	// X[vݒ
	SMCR = (1<<SE);			// X[v ACh[h

	// ^C}Q CTC
	TCCR2A = (1<<WGM21);
	TCCR2B = (1<<CS22)|(1<<CS21)|(1<<CS20);		// CTC vXP[ 1/1024
	OCR2A = (CPU_CLOCK/1024/100)-1;		// ^C} 100Hz
	TIMSK2 = (1<<OCIE2A);	// r`荞݋

	// d̓Jbg
	PRR = (1<<PRTIM0)|(1<<PRTIM1)|(1<<PRSPI)|(1<<PRUSART0);

	DIDR0 = (1<<SCL)|(1<<SDA)|(1<<BATT);	// BATTsfW^͋֎~

#else	// ATmega8
	// X[vݒ
	MCUCR = (1<<SE);			// X[v ACh[h

	TCCR2 = (1<<WGM21)|(1<<CS22)|(1<<CS21)|(1<<CS20);	// CTC, 1/1024
	OCR2 = (CPU_CLOCK/1024/100)-1;		// ^C} 100Hz
	TIMSK = (1<<OCIE2);		// r荞݋
#endif

	ACSR = (1<<ADC);	// Rp[^dOFF

	ADMUX = (1<<REFS1)|(1<<REFS0)|BATT;			// d 1.1V, A/D ch

	sei();		// 荞݋

	// 0.1bԑ҂
	wait_timer(10);

	// ddႢꍇ͋NȂ
	while(check_batt()){
		wait_timer(1);
	}

	PORTD |= (1<<POWER);	// dێ	

	// JP2-1 ON ŏɃm
	R1Default = R1_DEFAULT;
	if (!(PINC & (1<<PC1)))	R1Default |= R1_MONO;

	// EEPROMÕ`lԍ擾
	FMch = eeprom_read_byte((uint8_t *)EEPROM_CH);
	if (FMch > CH_N){
		FMch = 0;
	}
	LEDdata = 1 << (FMch + 1);	// LED_
	
	wait_timer(50);		// d肷܂ő҂
	
	// TWI Zbg
	TWI_start();
	TWI_send(0xFF);
	TWI_start();
	TWI_stop();

	// AR1000 WX^l
	write_AR1000(R0, R0_DEFAULT | R0_XO_EN);	// ŏ͕K ENABLE=OFF
	write_AR1000(R1, R1Default | R1_HMUTE);	// ~[g
	write_AR1000(R2, R2_DEFAULT);
	write_AR1000(R3, R3_DEFAULT | R3_VOLUME);
	write_AR1000(R4, 0x0400);
	write_AR1000(R5, 0x28AA);
	write_AR1000(R6, 0x4400);
	write_AR1000(R7, 0x1EE7);
	write_AR1000(R8, 0x7141);
	write_AR1000(R9, 0x007D);
	write_AR1000(R10, 0x82CE);
	write_AR1000(R11, 0x4F55);
	write_AR1000(R12, 0x970C);
	write_AR1000(R13, 0xB845);
	write_AR1000(R14, 0xFC2D);
	write_AR1000(R15, 0x8097);
	write_AR1000(R16, 0x04A1);
	write_AR1000(R17, 0xDF6A);

	write_AR1000(R0, R0_DEFAULT | R0_XO_EN | R0_ENABLE);	// ENABLE
	wait_timer(250);	// 2.5b WAIT	ENABLE͏\Ȏԑ҂Kv炵
	for (;;){
		wait_timer(10);	// 100ms WAIT
		if (read_AR1000(STATUS) & STATUS_STC)	break;	// STC=H ҂
	}

	write_AR1000(R1, R1Default);	// ~[g

	// SW2SW4Ȃdꂽgݒ胂[h
	if ((SwFlag & (SW_1|SW_2|SW_3|SW_4)) == (SW_2|SW_4)){
		scan_wave();	// gXLݒ
	}

	LEDdata = 1 << (FMch + 1);
	freq = get_freq(FMch);	// EEPROMg擾
	set_freq(freq);			// gݒ

	// XCb`܂ő҂
	while (SwFlag){
		wait_timer(1);
	}

	off_count = 0;
	for(;;){

		wait_timer(1);	// 10ms SLEEP

		if (SleepFlag && (SleepTimerM >= SLEEP_MIN)){
			power_off();	// dOFF
		}
		
		if (SwFlag & SW_POWER){

			// POWER SW
			if (++off_count >= 20){		// 200ms͖
				LEDdata = 0;	// LED
				// {^܂ő҂
				while(SwFlag & SW_POWER){
					wait_timer(1);
					if (off_count < 255)	++off_count;
					if (off_count == 200){	// 2bŃX[v[h
						FlashLED = FMch + 1;	// LED_
					}
				}
				if (off_count < 200){	// 2b
					power_off();		// dOFF
				} else {
					// X[v[hݒ
					SleepTimerDS = SleepTimerS = SleepTimerM = 0;
					SleepFlag = 1;
					FlashLED = 0;
					LEDdata = 1 << (FMch + 1);
				}
			}
		} else if (SwFlag){
			// 1?4 SW
			FMch = 0;
			while(!(SwFlag & SW_1)){
				SwFlag >>= 1;
				FMch++;
			}
			LEDdata = 1 << (FMch + 1);			// LED_
			freq = get_freq(FMch);
			set_freq(freq);
			off_count = 0;

			// XCb`܂ő҂
			while(SwFlag){
				wait_timer(1);
			}
		} else {
			off_count = 0;
		}

		// obe[`FbN
		if (check_batt()){
			power_off();		// dOFF
		}
	}
}


//*******************************************************
//	drd`FbN
//*******************************************************
char check_batt(void){
	long l;

	cli();
	l = BattVolt;
	sei();
	l *= R8_VALUE + R9_VALUE;
	l /= R8_VALUE;
	l *= ADC_REF;
	l /= 1023;
	if (l < LOW_BATT){
		return 1;
	} else {
		return 0;
	}
}


//*******************************************************
//	dOFF
//*******************************************************
void power_off(void){

	if (eeprom_read_byte((uint8_t *)EEPROM_CH) != FMch){
		eeprom_write_byte((uint8_t *)EEPROM_CH, FMch);	// `lԍۑ
	}
	cli();
	LEDdata = 0;
	PORTD &= ~(1<<POWER);		// dێ
	for (;;){		// d؂
		WDR;
		SLEEP;
	}
}


//*******************************************************
//	gXLݒ
//*******************************************************
void scan_wave(void){
	uint	memory[CH_N];
	uchar	ch;
	uint	freq;
	uchar	sw;
	uchar	t;
	uchar	r;

	// XCb`܂ LED1_
	FlashLED = 1;
	while(SwFlag){
		wait_timer(10);
	}
	wait_timer(10);	// Ô

	// ݒ`l{^͑҂
	t = 100;		// 10b
	while(!(sw = SwFlag)){
		wait_timer(10);
		if (--t == 0){
			FlashLED = 0;	// Ԑ؂
			return;
		}
	}

	if (sw & SW_POWER){
		// POWER SW̏ꍇ͑S`lXL
		memset(memory, 0xFF,  sizeof(memory));
		freq = 760;
		ch = 0;
	} else {
		// `lSW̏ꍇ͂̃`lf[^
		// ̃`lgl߂
		for (ch = 0; ch < CH_N; ch++){
			memory[ch] = get_freq(ch);
		}
		for (ch = 0; ch < CH_N; ch++){
			sw >>= 1;
			if (sw & 1){
				break;
			}
		}
		for (; ch < (CH_N - 1); ch++){
			memory[ch] = memory[ch + 1];
		}
		if (memory[ch] >= 1080)	memory[ch] = 760;
		memory[ch] += 3;
		freq = memory[ch];
	}
	FlashLED = ch + 1;

	// XCb`܂ő҂
	while(SwFlag){
		wait_timer(10);
	}
	wait_timer(10);	// Ô
	
	// gXLJn
	while(ch < CH_N){

		while (!(r = check_freq(freq))){
			if (SwFlag & SW_POWER)	break;	// POWER SWŒf
			freq++;
			if (freq > 1080)	break;
		}

		if (!r)	break;

		memory[ch++] = freq;

		if (SwFlag & SW_POWER)	break;	// POWER SWŒf

		wait_timer(50);

		FlashLED++;			// LED_
		freq += 3;			// 300kHzXLbv
		if (freq > 1080){
			break;
		}
	}
	FlashLED = 0;

	if (!SwFlag){
		// ݒg
		for (ch = 0; ch < CH_N; ch++){
			if (eeprom_read_word((uint16_t *)(EEPROM_FREQ_START + (ch * 2)))
				!= memory[ch]){
				eeprom_write_word((uint16_t *)((EEPROM_FREQ_START + (ch * 2))),
								  memory[ch]);
			}
		}
	}
}


//*******************************************************
//	gݒ
//*******************************************************
void set_freq(uint freq){

	if ((freq < 760)||(freq > 1080))	return;

	freq -= 690;
	write_AR1000(R2, (R2_DEFAULT & ~(R2_TUNE|R2_CHAN)) + freq);
	write_AR1000(R2, (R2_DEFAULT & ~(R2_TUNE|R2_CHAN)) + freq + R2_TUNE);

	for (;;){
		wait_timer(2);
		if (read_AR1000(STATUS) & STATUS_STC)	break;	// STC
	}
}


//*******************************************************
//	`lԍg擾
//*******************************************************
uint get_freq(uchar ch){
	uint	freq;

	if (ch >= CH_N)	return 0xFFFF;

	// EEPROMgl(100kHzP)ǂݍ
	freq = eeprom_read_word((uint16_t *)(EEPROM_FREQ_START + (ch * 2)));

	return freq;
}


//*********************************************************
//	̎gɓdg邩ׂ
//*********************************************************
uchar check_freq(uint freq){
	uint	rssi;
	uint	if_cnt;
	uint	reg;
	uchar	i;

	set_freq(freq);		// gݒ

	wait_timer(8);		// Ƒ҂

	rssi = 0;
	if_cnt = 0;
	for (i = 0; i < 4; i++){
		wait_timer(2);		// Ƒ҂
		reg = read_AR1000(RSSI);
		rssi += (uchar)(reg >> 9);
		if_cnt += reg & RSS_IF_CNT;
	}
	rssi += 2;
	rssi /= 4;
	if_cnt += 2;
	if_cnt /= 4;

	if ((rssi >= RSSI_MIN_LEVEL)&&
		(if_cnt >= 223)&&
		(if_cnt <= 277)){
		return 1;
	} else {
		return 0;
	}
}


//*******************************************************
//	AR1000 WX^(TWI)
//*******************************************************
void write_AR1000(uchar adrs, ushort data){

	TWI_start();
	TWI_send(0x20);			// X[uAhX + Write
	TWI_send(adrs);
	TWI_send(data >> 8);
	TWI_send(data & 0x00FF);
	TWI_stop();
}


//*******************************************************
//	AR1000 WX^ǂݍ(TWI)
//*******************************************************
ushort read_AR1000(uchar adrs){
	ushort	data;

	TWI_start();
	TWI_send(0x20);			// X[uAhX + Write
	TWI_send(adrs);
	TWI_start();
	TWI_send(0x20 + 1);		// X[uAhX + Read
	data = TWI_recv(0);		// ACK
	data <<= 8;
	data |= TWI_recv(1);	// NACK
	TWI_stop();

	return data;
}


//*******************************************************
//	Timerpԑ҂
//*******************************************************
void wait_timer(uchar t){

	Timer = t;
	while(Timer){
		WDR;
		SLEEP;
	}
}


//*******************************************************
//	TWI Start
//*******************************************************
char TWI_start(void){
	uint	i;

	TWBR = 32;
	TWSR = 0;		// TWI clock	8MHz / (16 + 2 * TWBR) =  100kHz

	TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);	// Jno
	for (i = 0; i < 10000; i++){
		if (TWCR & (1<<TWINT))	break;
	}

	if ((TWSR & TW_NO_INFO) != TW_START){
		return 1;	// ERROR
	} else {
		return 0;
	}
}


//*******************************************************
//	TWI Send
//*******************************************************
char TWI_send(uchar data){
	uint	i;

	TWDR = data;	// f[^
	TWCR = (1<<TWINT)|(1<<TWEN);
	for (i = 0; i < 10000; i++){
		if (TWCR & (1<<TWINT))	break;
	}

	if ((TWSR & TW_NO_INFO) != TW_MT_DATA_ACK){
		return 1;	// ERROR
	} else {
		return 0;
	}
}


//*******************************************************
//	TWI Recieve
//*******************************************************
uchar TWI_recv(uchar ack){
	uint	i;
	uchar	info;

	if (ack){
		TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWEA);
		info = TW_MR_DATA_ACK;
	} else {
		TWCR = (1<<TWINT)|(1<<TWEN);
		info = TW_MR_DATA_NACK;
	}

	for (i = 0; i < 10000; i++){
		if (TWCR & (1<<TWINT))	break;
	}

	if ((TWSR & TW_NO_INFO) != info){
		return 0;		// ERROR
	} else {
		return TWDR;	// f[^
	}
}


//*******************************************************
//	TWI Stop
//*******************************************************
void TWI_stop(void){

	TWCR = (1<<TWINT)|(1<<TWSTO)|(1<<TWEN);
}


